package org.colcon.burocrat.client;

import java.util.HashMap;
import java.util.Map;

import org.colcon.burocrat.api.client.IClient;
import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.IArgList;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.model.ArgList;
import org.colcon.burocrat.model.Model;

import com.google.gwt.event.dom.client.ChangeEvent;
import com.google.gwt.event.dom.client.ChangeHandler;
import com.google.gwt.user.client.ui.Grid;
import com.google.gwt.user.client.ui.HTML;

/**
 * An auto-generated data-entry form that consists of a set of BurocratFields, 
 * which allows users to enter data values for a given argument list 
 * 
 * @author Dan Vegeto
 *
 */
public class BurocratForm extends BurocratField implements ChangeHandler 
{
	private Grid grid;
	
	private IArgList initialArgs, generatedArgs;
	
	private Map<IArg, String> buffer;
	
	public BurocratForm(IClient client, IArg arg)
	{
		this(client, arg, null);
	}
	
	public BurocratForm(IClient client, IArgList argList)
	{
		this(client, null, argList);
	}
	
	public BurocratForm(IClient client, IArg arg, IArgList argList) 
	{
		super(client, arg);
		
		if(arg != null && argList != null)
			initialArgs = arg.subArgs().append(argList);
		else if(arg != null)
			initialArgs = arg.subArgs();
		else if(argList != null)
			initialArgs = argList;
	}
	
	public void init(IResponse data) 
	{
		grid = new Grid(0, 3);
		grid.setStyleName(CSS.FORM.getStyle());
		
		if(initialArgs != null)
			addFields(initialArgs);
		
		add(grid);
		
		generatedArgs = new ArgList();
		
		onChange(null);
	}
	
	public IArgList getArgs()
	{
		return initialArgs;
	}
	
	public HTML getLabel(int i)
	{
		return (HTML) grid.getWidget(i, 0);
	}
	
	public BurocratField getField(int i)
	{
		return (BurocratField) grid.getWidget(i, 1);
	}
	
	public HTML getStatus(int i)
	{
		return (HTML) grid.getWidget(i, 2);
	}
	
	public int getRow(IArg arg)
	{
		for(int i = 0; i < grid.getRowCount(); i++)
		{
			BurocratField field = getField(i);
			if(arg.equals(field.getArg()))
				return i;
		}
		
		return -1;
	}
	
	public BurocratField getField(IArg arg)
	{
		int i = getRow(arg);
		
		if(i == -1)
			return null;
		
		return getField(i);
	}
	
	private void addFields(IArgList argList)
	{
		for(int i = 0; i < argList.getArgs().length; i++)
		{
			IArg arg = argList.getArg(i);
			boolean required = argList.isRequired(i);
			
			if(getField(arg) == null)
				addField(arg, required);
		}
	}
	
	private void addField(IArg arg, boolean required)
	{
		int n = grid.getRowCount();
		
		HTML label = new HTML(arg.name());
		label.setStyleName(CSS.TEXT.getStyle());
		label.setTitle(arg.description());
		
		BurocratField field = getClient().widgets().buildField(getClient(), arg);
		
		field.addChangeHandler(this);
		
		HTML status = new HTML();
		
		if(required)
			status = getRequiredStatus();
		else
			status = getOptionalStatus();
		
		grid.resize(n + 1, grid.getColumnCount());
		grid.getRowFormatter().setStyleName(n, CSS.FORM_ROW.getStyle());
		
		grid.setWidget(n, 0, label);
		grid.setWidget(n, 1, field);
		grid.setWidget(n, 2, status);
	}
	
	public Map<IArg, String> getValueMap() 
	{
		Map<IArg, String> map = new HashMap<IArg, String>();
		
		for(int i = 0; i < grid.getRowCount(); i++)
		{
			BurocratField field = getField(i);
			String value = field.getValue();
			
			if(value != null && !value.isEmpty())
				map.put(field.getArg(), value);
		}
		
		return map;
	}
	
	public String getValue() 
	{
		return Model.instance.parser().composeArgs(getValueMap());
	}
	
	public String getValue(IArg arg)
	{
		return getValueMap().get(arg);
	}

	public void setValue(String value) 
	{
		Map<IArg, String> map = Model.instance.parser().parseArgs(value);
		
		setValue(map);
	}
	
	public void setValue(Map<IArg, String> map)
	{
		for(int i = 0; i < grid.getRowCount(); i++)
		{
			BurocratField field = getField(i);
			IArg arg = field.getArg();
			
			if(map.containsKey(arg))
			{
				field.setValue(map.get(arg));
				
				// is this necessary?
				onChange(null);
			}
		}
	}
	
	public HTML getRequiredStatus()
	{
		return new HTML("<img src=\"img/asterisk.gif\">");
	}
	
	public HTML getOptionalStatus()
	{
		return new HTML();
	}
	
	public HTML getValidStatus()
	{
		return new HTML("<img src=\"img/check.jpg\">");
	}
	
	public HTML getInvalidStatus()
	{
		return new HTML("<img src=\"img/x.jpg\">");
	}

	public void onChange(ChangeEvent event) 
	{
		Map<IArg, String> map = getValueMap();
		
		if(buffer != null && buffer.equals(map))
			return;
		else
			buffer = map;
		
		updateFields(map);
		updateStatuses(map);
		
		getChangeHandler().onChange(event);
	}
	
	private void updateStatuses(Map<IArg, String> values)
	{
		for(int i = 0; i < grid.getRowCount(); i++)
		{
			IArg arg = getField(i).getArg();
			
			HTML status = new HTML();
			
			if(values.containsKey(arg))
			{
				if(arg.validate(values.get(arg)).isSuccess())
					status = getValidStatus();
				else
					status = getInvalidStatus();
			}
			else
			{
				if(initialArgs.isRequired(arg) || generatedArgs.isRequired(arg))
					status = getRequiredStatus();
				else
					status = getOptionalStatus();
			}
			
			grid.setWidget(i, 2, status);
		}
	}
	
	private void updateFields(Map<IArg, String> values)
	{
		if(getArg().subArgs() == null)
			return;
		
		IArgList oldArgs = generatedArgs;
		IArgList newArgs = initialArgs.getArgDependency().getArgList(values);
		
		generatedArgs = newArgs;
		
		boolean different = oldArgs.size() != newArgs.size();
		
		for(int i = 0; i < oldArgs.size() && i < newArgs.size() && !different; i++)
		{
			if(!oldArgs.getArg(i).equals(newArgs.getArg(i)))
				different = true;
		}
		
		if(different)
		{
			for(int i = grid.getRowCount() - 1; i >= initialArgs.size(); i--)
				grid.removeRow(i);
			
			addFields(newArgs);
		}
		
		Map<IArg, String> map = initialArgs.getValueDependency().getValueMap(values);
		
		for(int i = 0; i < grid.getRowCount(); i++)
		{
			BurocratField field = getField(i); 
			IArg arg = field.getArg();
			
			if(map.containsKey(arg) && !field.getValue().equals(map.get(arg)))
				field.setValue(map.get(arg));
		}
	}
}
