package org.colcon.burocrat.client.list;

import java.util.Date;

import org.colcon.burocrat.api.client.IClient;
import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.ICommand;
import org.colcon.burocrat.api.model.IItem;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.client.CSS;
import org.colcon.burocrat.client.ui.BurocratPanel;
import org.colcon.burocrat.client.utils.TimeUtil;
import org.colcon.burocrat.model.Arg;
import org.colcon.burocrat.model.Cmd;
import org.colcon.burocrat.model.Command;

import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.user.client.ui.Grid;
import com.google.gwt.user.client.ui.HTML;
import com.google.gwt.user.client.ui.HTMLTable.Cell;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.ScrollPanel;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Widget;

/**
 * This is a panel used to display a list of items.
 * The panel displays a limited set of query results,
 * and allows the user to seek through pages of results.
 * It also allows the results to be sorted by different
 * properties (IArg values) of the items.
 * 
 * @author Dan Vegeto
 *
 */
public class BurocratList extends BurocratPanel implements ClickHandler
{
	public static final int
	
		ID_CHAR_LIMIT = 25,
		TEXT_CHAR_LIMIT = 140,
		PAGE_LIMIT = 8;
	
	protected static final IArg[]
			
		DEFAULT_COLUMNS = new IArg[]
		{
			Arg.ID
		};
	
	private HorizontalPanel header, hPanel;
	private Grid grid;
	private ScrollPanel scroll;
	
	private Label prevButton, nextButton;
	private Label pageLabel;
	
	private String path;
	
	private IArg[] columns;
	private String[] widths;
	
	private IItem[] items;
	
	private IArg sortColumn;
	private boolean sortAscending;
	private int offset;
	
	private boolean[] sortableColumns;
	
	public BurocratList(IClient client, String path)
	{
		this(client, path, DEFAULT_COLUMNS);
	}
	
	public BurocratList(IClient client, String path, IArg[] columns)
	{
		this(client, path, columns, null);
	}
	
	public BurocratList(IClient client, String path, IArg[] columns, String[] widths)
	{
		this(client, path, columns, widths, null);
	}
	
	public BurocratList(IClient client, String path, IArg[] columns, String[] widths, boolean[] sortableColumns) 
	{
		super(client);
		
		this.path = path;
		this.columns = columns;
		this.widths = widths;
		
		sortColumn = columns[0];
		sortAscending = false;
		offset = 0;
		
		this.sortableColumns = sortableColumns;
	}
	
	public String getPath()
	{
		return path;
	}
	
	public int getOffset()
	{
		return offset;
	}
	
	public IArg getSortColumn()
	{
		return sortColumn;
	}
	
	public boolean getSortAscending()
	{
		return sortAscending;
	}
	
	public void setSortAscending(boolean ascending)
	{
		sortAscending = ascending;
	}
	
	public ICommand getRequest()
	{
		// 1 is added to limit to check for next page
		
		return new Command(
				Arg.CMD.put(Cmd.QUERY),
				Arg.PATH.put(path),
				Arg.LIMIT.put(PAGE_LIMIT + 1),
				Arg.OFFSET.put(offset),
				Arg.SORT.put(sortColumn),
				Arg.ASCENDING.put(Boolean.toString(sortAscending)));
	}
	
	public void init(IResponse data)
	{
		setStyleName(CSS.LIST.getStyle());
		
		header = new HorizontalPanel();
		hPanel = new HorizontalPanel();
		
		prevButton = new Label("<");
		pageLabel = new Label();
		nextButton = new Label(">");

		pageLabel.setStyleName(CSS.LIST_PAGE_LABEL.getStyle());
		
		prevButton.addStyleName(CSS.LIST_PAGE_BUTTON.getStyle());
		prevButton.addStyleName(CSS.CLICKABLE.getStyle());
		
		nextButton.addStyleName(CSS.LIST_PAGE_BUTTON.getStyle());
		nextButton.addStyleName(CSS.CLICKABLE.getStyle());
		
		prevButton.addClickHandler(this);
		nextButton.addClickHandler(this);
		
		hPanel.add(prevButton);
		hPanel.add(pageLabel);
		hPanel.add(nextButton);
		
		header.add(hPanel);
		
		grid = new Grid();
		grid.addClickHandler(this);
		grid.setWidth("100%");
		
		scroll = new ScrollPanel(grid);
		scroll.setHeight("100%");
		
		add(header);
		add(scroll);
		
		Object obj = data.getObject();
		
		if(obj instanceof IItem[])
			items = (IItem[]) obj;
		else if(obj instanceof IItem)
			items = new IItem[]{(IItem)obj};
		else
			items = null;
		
		if(items == null || items.length == 0)
		{
			Label label;
			if(items == null)
				label = new Label(obj.toString());
			else
				label = new Label("no results returned by this query");
			
			label.setStyleName(CSS.TEXT.getStyle());
			
			grid.resize(1, 1);
			grid.setWidget(0, 0, label);
			
			prevButton.setVisible(false);
			nextButton.setVisible(false);
			pageLabel.setVisible(false);
			
			return;
		}
		
		boolean atStart = offset == 0;
		boolean atEnd = items.length <= PAGE_LIMIT;
		
		prevButton.setVisible(!atStart);
		nextButton.setVisible(!atEnd);
		pageLabel.setVisible(!atStart || !atEnd);
		
		int count;
		if(items.length > PAGE_LIMIT)
			count = PAGE_LIMIT;
		else
			count = items.length;
		
		String start = (offset + 1) + "";
		String end = (offset + count) + "";
		
		if(count == 1)
			pageLabel.setText(start);
		else
			pageLabel.setText(start + " - " + end);
		
		initGrid(items);
	}
	
	protected void initGrid(IItem[] items)
	{
		if(items == null)
			add(new Label("failed to open container '" + path + "'"));
		else if(items.length == 0)
			add(new Label("no items in this container"));
		else
		{
			int rows;
			if(items.length > PAGE_LIMIT)
				rows = PAGE_LIMIT + 1;
			else
				rows = items.length + 1;
			
			grid.resize(rows, columns.length);
			
			initGridHeader();
			
			for(int i = 0; i < items.length && i + 1 < grid.getRowCount(); i++)
				initGridRow(i + 1, items[i]);
		}
	}
	
	private void initGridHeader()
	{
		addRowStyle(0, CSS.CLICKABLE.getStyle());
		
		for(int i = 0; i < columns.length; i++)
		{
			if(widths != null && widths[i] != null)
				setColumnWidth(i, widths[i]);
			
			String str = columns[i].name();
			
			if(sortColumn.equals(columns[i]))
			{
				if(sortAscending)
					str = "+ " + str + " +";
				else
					str = "- " + str + " -";
			}
			
			Label label = new Label(str);
			label.setStyleName(CSS.LIST_HEADER.getStyle());
			
			if(sortableColumns == null || sortableColumns[i])
				label.addStyleName(CSS.CLICKABLE.getStyle());
			
			grid.setWidget(0, i, label);
		}
	}
	
	protected void initGridRow(int row, IItem item)
	{
		addRowStyle(row, "clickable");
		addRowStyle(row, "list-row");
		
		for(int i = 0; i < columns.length; i++)
			grid.setWidget(row, i, buildGridCell(item, columns[i]));			
	}
	
	protected void addRowStyle(int row, String style)
	{
		grid.getRowFormatter().addStyleName(row, style);	
	}
	
	protected void removeRowStyle(int row, String style)
	{
		grid.getRowFormatter().removeStyleName(row, style);	
	}
	
	protected void removeRowStyle(String style)
	{
		for(int i = 0; i < grid.getRowCount(); i++)
			removeRowStyle(i, style);
	}
	
	protected void setColumnWidth(int column, String width)
	{
		if(grid != null && column < grid.getColumnCount())
			grid.getColumnFormatter().setWidth(column, width);
	}
	
	protected Widget buildGridCell(IItem item, IArg column)
	{
		Widget widget = new Label("null");
		
		if(item != null && column != null)
		{
			String str = item.getArg(column);
			
			if(column == Arg.CREATED)
			{
				long created = Long.valueOf(str);
				Date createdDate = new Date(created);
				str = TimeUtil.getDifference(createdDate, new Date());
			}
			
			if(str == null)
				widget = new HTML("");
			else
				widget = new HTML(str);
		}
		
		widget.setStyleName(CSS.LIST_CELL.getStyle());
		
		return widget;
	}
	
	protected void click(int row, int col)
	{
		if(row == 0)
		{
			if(sortableColumns != null && !sortableColumns[col])
				return;
			
			if(!sortColumn.equals(columns[col]))
			{
				sortColumn = columns[col];
				sortAscending = true;
			}
			else
			{
				sortAscending = !sortAscending;
			}
			
			offset = 0;
			
			refresh();
		}
		else
		{
			int i = row - 1;
			
			String path = items[i].getPath();
			getClient().show(path);
		}
	}
	
	public void onClick(ClickEvent event) 
	{
		if(event.getSource().equals(grid))
		{
			Cell cell = grid.getCellForEvent(event);
			int row = cell.getRowIndex();
			int col = cell.getCellIndex();
			
			click(row, col);
		}
		else if(event.getSource().equals(prevButton))
		{
			if(offset >= PAGE_LIMIT)
				offset -= PAGE_LIMIT;
			else
				offset = 0;
			
			refresh();
		}
		else if(event.getSource().equals(nextButton))
		{
			offset += PAGE_LIMIT;
			refresh();
		}
	}
}
