package org.colcon.burocrat.model;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.IArgDependency;
import org.colcon.burocrat.api.model.IArgValue;
import org.colcon.burocrat.api.model.IFlag;
import org.colcon.burocrat.api.model.IArgList;
import org.colcon.burocrat.api.model.IChoice;
import org.colcon.burocrat.api.model.ICmd;
import org.colcon.burocrat.api.model.IFieldType;
import org.colcon.burocrat.api.model.IItemType;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.api.utils.ArrayUtil;

/**
 * 
 * These are arguments, or properties, which are the keys 
 * in the map representations of all data
 * 
 * @author Dan Vegeto
 *
 */
public enum Arg implements IArg
{
	VALUE("an arbitrary string property"),
	TEXT("the main content of an item"),
	ID("an identifier unique within this item's parent container"),
	VERSION("the version of this data"),
	SUBJECT("a brief description of something"),
	
	ARG(FieldType.LIST_BOX, "the key of an item or command property")
	{
		public IChoice[] choices()
		{
			return Model.instance.getArgs();
		}
	},
	
	SORT(ARG, "the method of sorting for this query"),
	
	ASCENDING(FieldType.TEXT_BOX, "the direction of sorting for this query"),
	
	ARGS(FieldType.FORM, "a set of key-value pairs"),
	FILTERS(ARGS, "a set of filters for this query"),
	
	COMMAND(ARGS, "a request to access or modify data")
	{
		public IArgList	subArgs()
		{			
			return new ArgList(new IArg[]{CMD}, new IArgDependency()
			{
				public IArgList	getArgList(Map<IArg, String> input)
				{
					if(input.get(CMD) != null)
					{
						String cmd = input.get(CMD);
						ICmd cmdEnum = Model.instance.getCmd(cmd);
						
						return cmdEnum.argList();
					}
					
					return new ArgList();
				}
			});
		}
	},
	
	ITEM(ARGS, "a set of key-value pairs that defines a data item")
	{
		public IArgList	subArgs()
		{			
			return new ArgList(new IArg[]{TYPE},
					
				new IArgDependency()
				{
					public IArgList	getArgList(Map<IArg, String> input)
					{
						if(input.get(TYPE) != null)
						{
							String type = input.get(TYPE);
							IItemType typeEnum = Model.instance.getItemType(type);
							
							return typeEnum.getAllArgs();
						}
						
						return new ArgList();
					}
				}
			);
		}
	},
	
	PATH("a unique identifier of this item"),
	PARENT(PATH, "the path to this item's parent"),
	SOURCE(PATH, "the path to the source of this link"),
	TARGET(PATH, "the path to the target of this link"),
	AGENT(PATH, "the path of an entity that can execute commands"),
	
	EMAIL("an email address"),
	SENDER(EMAIL, "the email address of the sender"),
	RECIPIENT(EMAIL, "the email address of the recipient"),
	
	TIME(FieldType.TEXT_BOX, "a date and time as a long value"),
	CREATED(TIME, "the time when this item was created"),
	ACCESSED(TIME, "the time when this item was last accessed"),
	MODIFIED(TIME, "the time when this item was last modified"),
	
	NUMBER("a number"),
	LIMIT(NUMBER, "the maximum number of query results"),
	OFFSET(NUMBER, "the number of query results to skip"),
	
	BOOLEAN(FieldType.RADIO_BUTTONS, "a true or false value")
	{
		public IChoice[] choices()
		{
			return BooleanChoice.values();
		}
	},
	
	PUBLIC(BOOLEAN, "a flag that determines if this item can be seen publicly"),
	EMAIL_CONFIRMED(BOOLEAN, "true if this user has verified their email address"),
	
	ENCRYPTED(FieldType.PASSWORD_BOX, "an encrypted string", new IFlag[]{ArgFlag.ENCRYPTED}),
	PASSWORD(ENCRYPTED, "the encrypted password for this user"),
	CONFIRM_PASSWORD(PASSWORD, "a repeat of the user's password"),
	CONFIRMATION_CODE(ENCRYPTED, "encrypted code used to verify an email address"),
	
	TYPE(FieldType.LIST_BOX, "the type of this item")
	{
		public IChoice[] choices()
		{
			return Model.instance.getItemTypes();
		}
	},
	
	CMD(FieldType.LIST_BOX, "a type of command")
	{
		public IChoice[] choices()
		{
			return Model.instance.getCmds();
		}
	},
	
	HELP_CMD(CMD, "a type of command");
	
	private IArg parentArg;
	private IFieldType fieldType;
	private String description;
	
	private Set<IFlag> flags;
	
	private Arg(String description)
	{
		this(null, FieldType.TEXT_BOX, description);
	}
	
	private Arg(IArg parentArg, String description)
	{
		this(parentArg, FieldType.TEXT_BOX, description);
	}
	
	private Arg(IFieldType inputType, String description)
	{
		this(null, inputType, description);
	}
	
	private Arg(IArg parentArg, IFieldType fieldType, String description)
	{
		this(parentArg, fieldType, description, new IFlag[0]);
	}
	
	private Arg(IFieldType inputType, String description, IFlag[] flags)
	{
		this(null, inputType, description, flags);
	}
	
	private Arg(IArg parentArg, IFieldType fieldType, String description, IFlag[] flags)
	{
		this.parentArg = parentArg;
		this.fieldType = fieldType;
		this.description = description;
		
		this.flags = new HashSet<IFlag>();
		this.flags.addAll(Arrays.asList(flags));
	}
	
	public IArg parent()
	{
		return parentArg;
	}
	
	public IFieldType fieldType()
	{
		if(parentArg != null)
			return parentArg.fieldType();
		
		return fieldType;
	}
	
	public String description()
	{
		return description;
	}
	
	public IFlag[] flags()
	{
		if(parentArg != null)
			return ArrayUtil.append(parentArg.flags(), flags.toArray(new IFlag[0]));
		
		return flags.toArray(new IFlag[0]);
	}
	
	public boolean hasFlag(IFlag flag)
	{
		if(parentArg != null && parentArg.hasFlag(flag))
			return true;
		
		return flags.contains(flag);
	}
	
	public IChoice[] choices()
	{
		if(parentArg != null)
			return parentArg.choices();
		
		return null;
	}
	
	public IArgList subArgs()
	{
		if(parentArg != null)
			return parentArg.subArgs();
		
		return null;
	}
	
	public IArgList	subArgs(Map<IArg, String> input)
	{
		return null;
	}
	
	public IResponse validate(String value)
	{
		if(parentArg != null && !parentArg.validate(value).isSuccess())
			return new Response(false);
		
		if(value == null)
			return new Response(false);
		
		IChoice[] choiceArray = choices();
		
		if(choiceArray != null)
		{
			for(IChoice choice : choiceArray)
				if(choice.name().equalsIgnoreCase(value))
					return new Response(true);
		
			return new Response(false);
		}
		
		IArgList subArgs = subArgs();
		
		if(subArgs != null)
		{
			Map<IArg, String> valueMap = Model.instance.parser().parseArgs(value);
			
			return subArgs.validate(valueMap);
		}
		
		return new Response(true);
	}

	public Object parse(String value) 
	{
		if(parentArg != null)
			return parentArg.parse(value);
		
		return value;
	}
	
	public String getHelpString()
	{
		String str = name().toLowerCase();
		
		while(str.length() < 15)
			str += ' ';
		
		str += description();
		
		return str;
	}
	
	public IArgValue put(Object value)
	{
		if(value == null)
			return null;
		
		return new ArgValue(this, value.toString());
	}
	
	public static IArg[] getValues()
	{
		return (IArg[])values();
	}
}
