package org.colcon.burocrat.model;

import java.util.LinkedHashMap;
import java.util.Map;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.IArgDependency;
import org.colcon.burocrat.api.model.IArgList;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.api.model.IValueDependency;
import org.colcon.burocrat.api.utils.ArrayUtil;

/**
 * 
 * A set of arguments, each of which is either optional or required,
 * which can also have additional form logic (ArgDependency, ValueDependency)
 * 
 * @author Dan Vegeto
 *
 */
public class ArgList implements IArgList 
{
	private IArg[] args;
	private boolean[] required;
	
	private IArgDependency argDependency;
	private IValueDependency valueDependency;
	
	public ArgList()
	{
		this(new IArg[0]);
	}
	
	public ArgList(IArg... args)
	{
		this(args, new ArgDependency());
	}
	
	public ArgList(IArg[] requiredArgs, IArg[] optionalArgs)
	{
		this(requiredArgs, optionalArgs, new ArgDependency());
	}
	
	public ArgList(IArg[] requiredArgs, IArgDependency argDependency)
	{
		this(requiredArgs, new IArg[0], argDependency);
	}
	
	public ArgList(IArg[] requiredArgs, IArgDependency argDependency, IValueDependency valueDependency)
	{
		this(requiredArgs, new IArg[0], argDependency, valueDependency);
	}
	
	public ArgList(IArg[] requiredArgs, IArg[] optionalArgs, IArgDependency argDependency)
	{
		this(requiredArgs, optionalArgs, argDependency, new ValueDependency());
	}
	
	public ArgList(IArg[] requiredArgs, IArg[] optionalArgs, IArgDependency argDependency, IValueDependency valueDependency)
	{
		this(
				ArrayUtil.append(requiredArgs, optionalArgs), 
				new boolean[requiredArgs.length + optionalArgs.length],
				argDependency,
				valueDependency
			);
		
		for(int i = 0; i < requiredArgs.length; i++)
			required[i] = true;
	}
	
	public ArgList(IArg[] args, boolean[] required, IArgDependency argDependency, IValueDependency valueDependency)
	{
		this.args = args;
		this.required = required;
		this.argDependency = argDependency;
		this.valueDependency = valueDependency;
	}
	
	public IArg[] getArgs()
	{
		return args;
	}
	
	public boolean[] getRequired()
	{
		return required;
	}
	
	public IArgDependency getArgDependency()
	{
		return argDependency;
	}
	
	public IValueDependency getValueDependency()
	{
		return valueDependency;
	}
	
	public int size()
	{
		return args.length;
	}
	
	public IArg getArg(int i)
	{
		return args[i];
	}
	
	public boolean isRequired(int i)
	{
		return required[i];
	}
	
	public int getArgIndex(IArg arg)
	{
		for(int i = 0; i < size(); i++)
			if(getArg(i).equals(arg))
				return i;
		
		return -1;
	}
	
	public boolean isRequired(IArg arg)
	{
		int i = getArgIndex(arg);
	
		return i >= 0 && isRequired(i);
	}
	
	public IResponse validate(Map<IArg, String> map)
	{
		for(int i = 0; i < args.length; i++)
		{
			if(isRequired(i))
			{
				if(!map.containsKey(args[i]))
					return new Response(false, "missing required argument '" + args[i] + "'");
				if(!args[i].validate(map.get(args[i])).isSuccess())
					return new Response(false, "invalid argument '" + args[i] + "'");
			}
			else
			{
				if(map.containsKey(args[i]) && !args[i].validate(map.get(args[i])).isSuccess())
					return new Response(false, "invalid argument '" + args[i] + "'");
			}
		}
		
		IArgList generatedArgs = argDependency.getArgList(map);
		
		if(generatedArgs.size() > 0)
			return argDependency.getArgList(map).validate(map);
		
		return new Response(true);
	}
	
	public Map<IArg, String> withArgs(Object... values)
	{
		Map<IArg, String> map = new LinkedHashMap<IArg, String>();
		
		for(int i = 0; i < values.length && i < size(); i++)
			map.put(getArg(i), values[i].toString());
	
		if(values.length > size())
		{
			IArgList subList = getArgDependency().getArgList(map);
		
			Object[] subListValues = ArrayUtil.subArray(values, size(), size() + subList.size());
			
			map.putAll(subList.withArgs(subListValues));
		}
		
		return map;
	}
	
	public IArgList append(final IArgList argList)
	{
		return new ArgList(ArrayUtil.append(args, argList.getArgs()), ArrayUtil.append(required, argList.getRequired()),
				
			new IArgDependency()
			{
				public IArgList getArgList(Map<IArg, String> input) 
				{
					return getArgDependency().getArgList(input).append(argList.getArgDependency().getArgList(input));
				}			
			},
			
			new IValueDependency()
			{
				public Map<IArg, String> getValueMap(Map<IArg, String> input) 
				{
					Map<IArg, String> map = getValueDependency().getValueMap(input);
					
					map.putAll(argList.getValueDependency().getValueMap(input));
					
					return map;
				}			
			}
		);
	}
}
