package org.colcon.burocrat.model;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.IArgList;
import org.colcon.burocrat.api.model.ICmd;
import org.colcon.burocrat.api.model.IFlag;
import org.colcon.burocrat.api.model.IResponse;

/**
 * 
 * These are definitions of server commands, each with an argument list 
 * and a set of command flags. The commands are implemented in the server code. 
 * 
 * @author Dan Vegeto
 *
 */
public enum Cmd implements ICmd
{
	HELP("request a list of commands and their descriptions", 
			new ArgList(
					new IArg[]{}, 
					new IArg[]{Arg.HELP_CMD})),
	
	DO_NOTHING("does nothing"),
	
	AUTHENTICATE("validate the client's user credentials",
			new ArgList(Arg.ID, Arg.PASSWORD),
			new IFlag[]{CmdFlag.OPEN}),
	
	VALIDATE("validate a command without executing it", 
			new ArgList(Arg.COMMAND),
			new IFlag[]{CmdFlag.OPEN}),
	
	GET("request the data for a single item", 
			new ArgList(Arg.PATH),
			new IFlag[]{CmdFlag.OPEN}),
			
	QUERY("request the data for a set of items", 
			new ArgList(
					new IArg[]{}, 
					new IArg[]{Arg.PATH, Arg.LIMIT, Arg.OFFSET, Arg.SORT, Arg.ASCENDING, Arg.FILTERS}),
					new IFlag[]{CmdFlag.OPEN}),
			
	POST("add a new data item to the database", 
			new ArgList(Arg.ITEM),
			new IFlag[]{CmdFlag.PUBLIC}),
			
	PUT("modify existing data in the database", 
			new ArgList(Arg.PATH, Arg.ARG, Arg.VALUE),
			new IFlag[]{CmdFlag.PUBLIC}),
			
	DELETE("remove data from the database", 
			new ArgList(Arg.PATH),
			new IFlag[]{CmdFlag.PUBLIC}),
			
	QUERY_LINKS("request the links for a given source or target",
			new ArgList(
					new IArg[]{}, 
					new IArg[]{Arg.TYPE, Arg.SOURCE, Arg.TARGET, Arg.LIMIT, Arg.OFFSET, Arg.SORT, Arg.ASCENDING}),
			new IFlag[]{CmdFlag.OPEN}),
			
	QUERY_LINK_SOURCES("request the sources of links with a given target",
			new ArgList(
					new IArg[]{}, 
					new IArg[]{Arg.TYPE, Arg.TARGET, Arg.LIMIT, Arg.OFFSET, Arg.SORT, Arg.ASCENDING}),
			new IFlag[]{CmdFlag.OPEN}),
	
	QUERY_LINK_TARGETS("request the targets of links with a given source",
			new ArgList(
					new IArg[]{}, 
					new IArg[]{Arg.TYPE, Arg.SOURCE, Arg.LIMIT, Arg.OFFSET, Arg.SORT, Arg.ASCENDING}),
			new IFlag[]{CmdFlag.OPEN}),
			
	EMAIL("send an email",
			new ArgList(Arg.RECIPIENT, Arg.SUBJECT, Arg.TEXT)),
			
	REGISTER("register a user",
			new ArgList(Arg.EMAIL, Arg.ID, Arg.PASSWORD, Arg.CONFIRM_PASSWORD),
			new IFlag[]{CmdFlag.PUBLIC, CmdFlag.OPEN}),
			
	CONFIRM_EMAIL("verifies a user's email address",
			new ArgList(Arg.ID, Arg.CONFIRMATION_CODE),
			new IFlag[]{CmdFlag.OPEN})
			
	;
	
	private String description;
	private IArgList argList;
	private Set<IFlag> flags;
	
	private Cmd(String description)
	{
		this(description, new ArgList(), new IFlag[0]);
	}
	
	private Cmd(String description, IArgList argList)
	{
		this(description, argList, new IFlag[0]);
	}
	
	private Cmd(String description, IArgList argList, IFlag[] flags)
	{
		this.description = description;
		this.argList = argList;
		
		this.flags = new HashSet<IFlag>();
		this.flags.addAll(Arrays.asList(flags));
	}
	
	public String description()
	{
		return description;
	}
	
	public IArgList argList()
	{
		return argList;
	}
	
	public IFlag[] flags()
	{
		return flags.toArray(new IFlag[0]);
	}
	
	public boolean hasFlag(IFlag flag)
	{
		return flags.contains(flag);
	}
	
	public IResponse validate(String agent, Map<IArg, String> args)
	{
		return argList().validate(args);
	}
	
	public String getNews(String agent, Map<IArg, String> args)
	{
		return "#" + agent + " executed  command '" + name() + "'";
	}
	
	public String getHelpString()
	{
		String str = name().toLowerCase();
		
		while(str.length() < 15)
			str += ' ';
		
		str += description();
		
		return str;
	}
	
	public Map<IArg, String> withArgs(Object... values)
	{
		Map<IArg, String> map = argList().withArgs(values);
		
		map.put(Arg.CMD, name());
		
		return map;
	}
	
	public static ICmd[] getValues()
	{
		return (ICmd[])values();
	}
}
