package org.colcon.burocrat.model;

import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.IArgList;
import org.colcon.burocrat.api.model.IFlag;
import org.colcon.burocrat.api.model.IItem;
import org.colcon.burocrat.api.model.IItemType;
import org.colcon.burocrat.api.utils.ArrayUtil;
import org.colcon.burocrat.api.utils.StringUtil;

/**
 * 
 * These are definitions of item types, each with an argument list, 
 * a set of item types for subcontainers, a set of item type flags, 
 * and an init(IItem) method that gets called whenever an item of 
 * this type is posted to the server
 * 
 * @author Dan Vegeto
 *
 */
public enum ItemType implements IItemType
{
	ITEM("items", "the root type of all items", null, new ArgList(), new IItemType[0], new IFlag[0])
	{
		public void init(IItem instance) 
		{
			super.init(instance);
			
			instance.setArg(Arg.CREATED, Long.toString(new Date().getTime()));
			
			if(!instance.hasArg(Arg.PUBLIC))
				instance.setArg(Arg.PUBLIC, Boolean.TRUE.toString());
		}
	},
	
	CONTAINER("containers", "an item that contains other items", ITEM, new ArgList(Arg.PARENT), new IItemType[0], new IFlag[0]),
	
	NEWS("news", "a record of a command that was executed", ITEM, new ArgList(Arg.AGENT, Arg.COMMAND, Arg.PUBLIC), new IItemType[0], new IFlag[0])
	{
		public void init(IItem instance)
		{
			super.init(instance);
			instance.setArg(Arg.PARENT, plural());
		}
	},
	
	LINK("links", "a directional relation between two items", ITEM, new ArgList(Arg.SOURCE, Arg.TARGET), new IItemType[0], new IFlag[0])
	{
		public void init(IItem instance)
		{
			super.init(instance);
			instance.setArg(Arg.PARENT, plural());
		}
	},
	
	NOTIFICATION("notifications", "a link between news and an item", LINK, new ArgList(), new IItemType[0], new IFlag[0]),
	
	METADATA("metadatas", "information about the site's data", ITEM, new ArgList(Arg.VERSION), new IItemType[0], new IFlag[0])
	{
		public String generateId(IItem instance) 
		{
			return name().toLowerCase();
		}
	},
	
	POWER("powers", "a permission to execute a command with certain restrictions", ITEM, new ArgList(Arg.COMMAND), new IItemType[0], new IFlag[0]),
	
	AGENT("agents", "an entity that can execute commands", ITEM, new ArgList(), new IItemType[]{POWER}, new IFlag[0]),
	
	USER("users", "a user's personal information", AGENT, new ArgList(Arg.EMAIL, Arg.ID, Arg.PASSWORD), new IItemType[0], new IFlag[0])
	{
		private final int CONFIRMATION_CODE_LENGTH = 32;
		
		public void init(IItem instance)
		{
			super.init(instance);
			instance.setArg(Arg.PARENT, plural());
			
			String confirmationCode = StringUtil.randomString(CONFIRMATION_CODE_LENGTH);
			instance.setArg(Arg.CONFIRMATION_CODE, confirmationCode);
			
			if(!instance.hasArg(Arg.EMAIL_CONFIRMED))
				instance.setArg(Arg.EMAIL_CONFIRMED, Boolean.toString(false));
		}
	}
	;
	
	private String plural, description;
	private IItemType parent;
	
	private IArgList args;
	private IItemType[] containerTypes;
	private Set<IFlag> flags;
	
	private ItemType(String plural, String description, IItemType parent, IArgList args, IItemType[] containers, IFlag[] flags)
	{
		this.plural = plural;
		this.description = description;
		this.parent = parent;
		this.args = args;
		this.containerTypes = containers;
		
		this.flags = new HashSet<IFlag>();
		this.flags.addAll(Arrays.asList(flags));
	}
	
	public String plural()
	{
		return plural;
	}
	
	public String description()
	{
		return description;
	}
	
	public IItemType parent()
	{
		return parent;
	}
	
	public IArgList args()
	{
		return args;
	}
	
	public IArgList getAllArgs()
	{
		if(parent != null)
			return parent.getAllArgs().append(args());
		
		return args();
	}
	
	public IItemType[] containerTypes()
	{
		return containerTypes;
	}
	
	public IItemType[] getAllContainerTypes()
	{
		if(parent != null)
			return ArrayUtil.append(parent.getAllContainerTypes(), containerTypes());
		
		return containerTypes();
	}
	
	public IItem[] getInitItems(IItem instance) 
	{
		return new IItem[0];
	}
	
	public IItem[] getAllInitItems(IItem instance) 
	{
		if(parent != null)
			return ArrayUtil.append(parent.getAllInitItems(instance), getInitItems(instance));
		
		return new IItem[0];
	}
	
	public IFlag[] flags()
	{
		if(parent() != null)
			return ArrayUtil.append(parent().flags(), flags.toArray(new IFlag[0]));
		
		return flags.toArray(new IFlag[0]);
	}
	
	public boolean hasFlag(IFlag flag)
	{
		if(parent() != null && parent().hasFlag(flag))
			return true;
		
		return flags.contains(flag);
	}

	public void init(IItem instance) 
	{
		if(parent != null)
			parent.init(instance);
		
		if(instance.getId() == null)	
			instance.setArg(Arg.ID, generateId(instance));
		
		if(instance.getParent() == null)
			instance.setArg(Arg.PARENT, Constants.ROOT);
	}
	
	public String generateId(IItem instance)
	{
		if(parent != null)
			return parent.generateId(instance);
		
		return null;
	}
	
	public Map<IArg, String> withArgs(Object... values)
	{
		Map<IArg, String> map = getAllArgs().withArgs(values);
		
		map.put(Arg.TYPE, name());
		
		return map;
	}
	
	public String getInstancePath(String id)
	{
		return Model.instance.parser().composePath(plural(), id);
	}
	
	public static IItemType[] getValues()
	{
		return (IItemType[])values();
	}
}


