package org.colcon.burocrat.model;

import java.util.LinkedHashMap;
import java.util.Map;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.ICmd;
import org.colcon.burocrat.api.model.ICommand;
import org.colcon.burocrat.api.model.IItem;
import org.colcon.burocrat.api.model.ILink;
import org.colcon.burocrat.api.model.IParser;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.api.model.IResponseType;

/**
 * 
 * This class allow Burocrat to convert between string and object representations of data
 * 
 * @author Dan Vegeto
 *
 */
public class Parser implements IParser
{
	public final String
	
		PATH_DELIMETER = "/",
		ITEM_DELIMETER = ";\n",
		ARG_DELIMETER = ",",
		RESPONSE_DELIMETER = ":",
		LINK_DELIMETER = "?";
	
	public Map<IArg, String> stringsToArgs(Map<String, String> map)
	{
		if(map == null)
			return null;
		
		Map<IArg, String> args = new LinkedHashMap<IArg, String>();
		
		for(String key : map.keySet())
		{
			IArg arg = Model.instance.getArg(key);
			if(arg != null)
				args.put(arg, map.get(key));
		}
		
		return args;
	}
	
	public Map<String, String> argsToStrings(Map<IArg, String> args)
	{
		if(args == null)
			return null;
		
		Map<String, String> map = new LinkedHashMap<String, String>();
		
		for(IArg key : args.keySet())
			map.put(key.name(), args.get(key));
		
		return map;
	}
	
	public Map<IArg, String> parseArgs(String str)
	{
		return stringsToArgs(parse(str));
	}
	
	public String composeArgs(Map<IArg, String> map)
	{
		return compose(argsToStrings(map));
	}
	
	public IItem parseItem(String str)
	{
		return new Item(parseArgs(str));
	}
	
	public String composeItem(IItem item)
	{
		return compose(argsToStrings(item.getArgs()));
	}
	
	public ICommand parseCommand(String str)
	{
		ICommand command = new Command(parseArgs(str));
	
		// add '-cmd' if not in command
		if(command.getCmd() == null)
		{
			String cmd = str.split(" ")[0];
			ICmd cmdEnum = Model.instance.getCmd(cmd);
			
			if(cmdEnum != null)
				command.setArg(Arg.CMD, cmdEnum.name());
		}
		
		return command;
	}
	
	public String composeCommand(ICommand command)
	{
		return composeArgs(command.getArgs());
	}
	
	public IResponse parseResponse(String str)
	{
		boolean multiline = str.split("\n").length > 1;
		String delim = multiline ? "\n" : ": ";
		
		String type = str.split(delim)[0];
		String message = str.substring(type.length() + delim.length());
		
		IResponseType typeEnum = ResponseType.valueOf(type);
		
		return new Response(typeEnum, message);
	}
	
	public String composeResponse(IResponse response)
	{
		String message = response.getMessage();
		boolean multiline = message.split("\n").length > 1;
		String delim = multiline ? "\n" : ": ";
		
		String type = response.getType().name();
		
		return type + delim + message;
	}
	
	public ILink parseLink(String str)
	{
		String[] array = str.split(LINK_DELIMETER);
		
		String path = array[0];
		String args = "";
		
		if(array.length > 1)
			args = array[1];
		
		Map<IArg, String> map = parseArgs(args);
		
		return new Link(path, map);
	}
	
	public String composeLink(ILink link)
	{
		String path = link.getPath();
		String args = composeArgs(link.getArgs());
		
		if(args == null || args.isEmpty())
			return path;
		
		return path + LINK_DELIMETER + args;
	}
	
	public IItem[] parseItemArray(String str)
	{
		if(str == null)
			return null;
		
		if(str.isEmpty())
			return new IItem[0];
		
		if(!str.contains(ITEM_DELIMETER))
			return new IItem[]{parseItem(str)};
		
		String[] array = str.split("\\s*" + ITEM_DELIMETER + "\\s*");
		IItem[] items = new IItem[array.length];
		
		for(int i = 0; i < array.length; i++)
			items[i] = parseItem(array[i]);
		
		return items;
	}
	
	public String composeItemArray(IItem... array)
	{
		String str = "";
		
		for(IItem item : array)
			str += composeItem(item) + ITEM_DELIMETER;
		
		return str;
	}
	
	public String[] parseArgArray(String str)
	{
		if(str == null || str.isEmpty())
			return null;
		
		if(!str.contains(ARG_DELIMETER))
			return new String[]{str};
		
		String[] array = str.split("\\s*" + ARG_DELIMETER + "\\s*");
		
		return array;
	}
	
	public String composeArgArray(String... array)
	{
		String str = "";
		
		for(String value : array)
			str += value + ARG_DELIMETER + ' ';
		
		return str;
	}
	
	public String[] parsePath(String path)
	{
		if(path == null || path.isEmpty())
			return null;
		
		if(!path.contains(PATH_DELIMETER))
			return new String[]{path};
		
		return path.split(PATH_DELIMETER);
	}
	
	public String composePath(String... array)
	{
		if(array == null)
			return null;
		
		if(array.length == 0)
			return "";
		
		int start = 0;
		String first = array[0];
		
		// ignore 'root' at start of path
		if(first == null || first.equals(Constants.ROOT))
		{
			if(array.length == 1)
				return "";

			start = 1;
		}
		
		String str = array[start];
		
		for(int i = start + 1; i < array.length; i++)
			str += PATH_DELIMETER + array[i];
		
		return str;
	}
	
	public String pathToParent(String path)
	{
		String[] array = parsePath(path);
		
		if(array.length <= 1)
			return "";
		
		String[] parent = new String[array.length - 1];
		
		for(int i = 0; i < parent.length; i++)
			parent[i] = array[i];
		
		return composePath(parent);
	}
	
	public String pathToId(String path)
	{
		String[] array = parsePath(path);
		
		return array[array.length - 1];
	}
	
	public Map<String, String> parse(String str) 
	{
		if(str == null)
			return null;
		
		Map<String, String> args = new LinkedHashMap<String, String>();
		
		if(str.trim().isEmpty())
			return args;
		
		str += " ";
		
		String key = "";
		String value = "";
		int nested = 0;
		boolean parsingKey = true;
		
		for(int i = 0; i < str.length(); i++)
		{
			char c = str.charAt(i);
			
			if(c == '{')
			{
				nested++;
				
				if(nested == 1)
					continue;
			}
			else if(c == '}')
			{
				nested--;
				
				if(nested == 0)
					continue;
			}
			
			if(nested == 0)
			{
				if(c == ' ')
				{
					if(parsingKey)
						parsingKey = false;
					else if(key.trim().length() > 0)
					{
						key = key.trim();
						value = value.trim();
						
						if(args.containsKey(key))
							args.put(key, args.get(key) + "," + value);
						else
							args.put(key, value);
						
						key = "";
						value = "";
					}
				}
				else if(c == '-' && i+1 < str.length() && 
						!Character.isDigit(str.charAt(i+1)) && str.charAt(i+1) != ' ' &&
						(i == 0 || str.charAt(i-1) == ' ' || str.charAt(i-1) == ';' || str.charAt(i-1) == '{' || str.charAt(i-1) == '\n'))
				{
					key = "";
					value = "";
					parsingKey = true;
				}
				else
				{
					if(parsingKey)
						key += c;
					else
						value += c;
				}
			}
			else
			{
				value += c;
			}
		}
			
		return args;
	}

	public String compose(Map<String, String> map) 
	{
		String str = "";
		
		for(String key : map.keySet())
		{
			String value = map.get(key);
			
			if(value == null)
				continue;
			
			if(value.split(" ").length > 1)
				value = '{' + value + '}';
			
			str += '-' + key + ' ' + value + ' ';
		}
		
		str = str.trim();
		
		return str;
	}
}
