package org.colcon.burocrat.server;

import java.util.HashMap;
import java.util.Map;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.IArgList;
import org.colcon.burocrat.api.model.ICmd;
import org.colcon.burocrat.api.model.ICommand;
import org.colcon.burocrat.api.model.IItem;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.api.server.ICmdImpl;
import org.colcon.burocrat.model.Arg;
import org.colcon.burocrat.model.ArgFlag;
import org.colcon.burocrat.model.Constants;
import org.colcon.burocrat.model.Item;
import org.colcon.burocrat.model.ItemType;
import org.colcon.burocrat.model.Model;
import org.colcon.burocrat.model.Response;
import org.colcon.burocrat.model.ResponseType;

/**
 * 
 * These are the server-side implementations of the command types with the same names, 
 * each with validate and execute methods
 * 
 * @author Dan Vegeto
 *
 */
public enum CmdImpl implements ICmdImpl 
{
	/**
	 * Returns a help string that contains all commands and their arguments
	 */
	HELP
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String cmd = args.get(Arg.HELP_CMD);
			
			if(cmd != null)
			{
				ICmd cmdEnum = Model.instance.getCmd(cmd);
				
				String str = cmdEnum.getHelpString() + '\n';
				
				IArgList argList = cmdEnum.argList();
				
				for(int i = 0; i < argList.getArgs().length; i++)
				{
					if(argList.isRequired(i))
						str += "\nrequired:	" + argList.getArg(i).getHelpString();
					else
						str += "\noptional:	" + argList.getArg(i).getHelpString();
				}
				
				return new Response(str);
			}
			else
			{
				ICmd[] cmds = Model.instance.getCmds();
				
				String str = cmds[0].getHelpString();
				
				for(int i = 1; i < cmds.length; i++)
					str += "\n" + cmds[i].getHelpString();
				
				return new Response(str);
			}
		}
	},
	
	/**
	 * Does nothing
	 */
	DO_NOTHING
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			return new Response();
		}
	},
	
	/**
	 * Validates the given user credentials
	 */
	AUTHENTICATE
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String id = args.get(Arg.ID);
			String password = args.get(Arg.PASSWORD);
			
			boolean result = Server.instance.authenticator().authenticate(id, password);
			
			return new Response(result);
		}
	},
	
	/**
	 * Validates a command without executing it
	 */
	VALIDATE
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String command = args.get(Arg.COMMAND);
			ICommand commandObj = Model.instance.parser().parseCommand(command);
			
			return Server.instance.validate(agent, commandObj);
		}
	},
	
	/**
	 * Used to access individual items
	 */
	GET
	{
		public IResponse validate(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			
			IItem item = Server.instance.get(path);
			
			if(item == null)
				return new Response(false, "item '" + path + "' not found");
			
			return new Response(true);
		}
		
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			
			IItem item = Server.instance.get(path);
			
			String str = item.toString();
			
			item.access();
			
			Server.instance.put(item);
			
			return new Response(ResponseType.ITEM, str);
		}
	},
	
	/**
	 * Used to access items and perform queries
	 */
	QUERY
	{
		public IResponse validate(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			
			if(path != null && !path.equals(Constants.ROOT))
			{
				IItem item = Server.instance.get(path);
			
				if(item == null)
					return new Response(false, "parent item '" + path + "' not found");
			}
			
			return new Response(true);
		}
		
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			String limit = args.get(Arg.LIMIT);
			String offset = args.get(Arg.OFFSET);
			String sort = args.get(Arg.SORT);
			String ascending = args.get(Arg.ASCENDING);
			String filters = args.get(Arg.FILTERS);
			
			int limitInt = Integer.MAX_VALUE;
			if(limit != null)
				limitInt = Integer.valueOf(limit);
			
			int offsetInt = 0;
			if(offset != null)
				offsetInt = Integer.valueOf(offset);
			
			IArg sortArg = null;
			if(sort != null)
				sortArg = Model.instance.getArg(sort);
			
			boolean ascendingBool = false;
			if(ascending != null)
				ascendingBool = Boolean.valueOf(ascending);
			
			Map<IArg, String> filtersMap = new HashMap<IArg, String>();
			if(filters != null)
				filtersMap = Model.instance.parser().parseArgs(filters);
			
			IItem[] items = Server.instance.query(path, limitInt, offsetInt, sortArg, ascendingBool, filtersMap);
			
			String str = Model.instance.parser().composeItemArray(items);
			
			// update the ACCESSED property of query results
			for(IItem item : items)
			{
				item.access();
				Server.instance.put(item);
			}
			
			return new Response(ResponseType.ITEM_LIST, str);
		}
	},
	
	/**
	 * Used to create new items
	 */
	POST
	{
		public IResponse validate(String agent, Map<IArg, String> args)
		{
			String item = args.get(Arg.ITEM);
			
			Map<IArg, String> itemMap = Model.instance.parser().parseArgs(item);
			
			if(itemMap.containsKey(Arg.PARENT) && !itemMap.get(Arg.PARENT).equals(Constants.ROOT))
			{
				String parent = itemMap.get(Arg.PARENT);
				
				IItem parentItem = Server.instance.get(parent);
				
				if(parentItem == null)
					return new Response(false, "parent item '" + parent + "' not found");
			}
			
			if(itemMap.containsKey(Arg.ID))
			{
				String id = itemMap.get(Arg.ID);
				String parent = itemMap.get(Arg.PARENT);
				String path = Model.instance.parser().composePath(parent, id);
				
				IItem existingItem = Server.instance.get(path);
				
				if(existingItem != null)
					return new Response(false, "item with path '" + path + "' already exists");
			}
			
			return new Response(true);
		}
		
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String itemArgs = args.get(Arg.ITEM);
			
			IItem item = Model.instance.parser().parseItem(itemArgs);
			
			Server.instance.post(item);
			
			IItem postedItem = Server.instance.get(item.getPath());
			
			boolean result = postedItem != null && postedItem.equals(item);
			
			return new Response(result, getNews(agent, args));
		}
	},
	
	/**
	 * Used to modify existing items
	 */
	PUT
	{
		public IResponse validate(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			String arg = args.get(Arg.ARG);
			String value = args.get(Arg.VALUE);
			
			if(Server.instance.get(path) == null)
				return new Response(false, "item '" + path + "' not found");
			
			IArg argEnum = Model.instance.getArg(arg);
			if(argEnum == null)
				return new Response(false, "arg '" + arg + "' not found");
			
			if(!argEnum.validate(value).isSuccess())
				return new Response(false, "invalid value '" + value + "' for arg '" + arg + "'");
			
			return new Response(true);
		}
		
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			String arg = args.get(Arg.ARG);
			String value = args.get(Arg.VALUE);
			
			IItem item = Server.instance.get(path);
			IArg argEnum = Model.instance.getArg(arg);
			
			if(argEnum.hasFlag(ArgFlag.ENCRYPTED))
				value = Server.instance.encryptor().encrypt(value);
			
			item.setArg(argEnum, value);
			item.modify();
			
			Server.instance.put(item);
			
			IItem postedItem = Server.instance.get(item.getPath());
			
			boolean result = postedItem != null && postedItem.getArg(argEnum).equals(value); 
			
			return new Response(result, getNews(agent, args));
		}
	},
	
	/**
	 * Used to remove items
	 */
	DELETE
	{
		public IResponse validate(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			
			IItem item = Server.instance.get(path);
			
			if(item == null)
				return new Response(false, "item '" + path + "' not found");
			
			return new Response(true);
		}
		
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String path = args.get(Arg.PATH);
			
			Server.instance.delete(path);
			
			IItem postedItem = Server.instance.get(path);
			
			boolean result = postedItem == null;
			
			return new Response(result, getNews(agent, args));
		}
	},
	
	QUERY_LINKS
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			Map<IArg, String> filtersMap = new HashMap<IArg, String>();
			
			filtersMap.put(Arg.TYPE, args.get(Arg.TYPE));
			filtersMap.put(Arg.SOURCE, args.get(Arg.SOURCE));
			filtersMap.put(Arg.TARGET, args.get(Arg.TARGET));
			
			String filters = Model.instance.parser().composeArgs(filtersMap);
			
			Map<IArg, String> queryArgs = new HashMap<IArg, String>();
			
			queryArgs.put(Arg.PATH, ItemType.LINK.plural());
			queryArgs.put(Arg.LIMIT, args.get(Arg.LIMIT));
			queryArgs.put(Arg.OFFSET, args.get(Arg.OFFSET));
			queryArgs.put(Arg.SORT, args.get(Arg.SORT));
			queryArgs.put(Arg.ASCENDING, args.get(Arg.ASCENDING));
			queryArgs.put(Arg.FILTERS, filters);
			
			return QUERY.execute(agent, queryArgs);
		}
	},
	
	QUERY_LINK_SOURCES
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			IResponse response = QUERY_LINKS.execute(agent, args);
			
			IItem[] links = (IItem[]) response.getObject();
			IItem[] sources = new IItem[links.length];
			
			for(int i = 0; i < links.length; i++)
			{
				String path = links[i].getArg(Arg.SOURCE);
				sources[i] = Server.instance.get(path);
			}
			
			String str = Model.instance.parser().composeItemArray(sources);
			
			// update the ACCESSED property of query results
			for(IItem item : sources)
			{
				item.access();
				Server.instance.put(item);
			}
			
			return new Response(ResponseType.ITEM_LIST, str);
		}
	},
	
	QUERY_LINK_TARGETS
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			IResponse response = QUERY_LINKS.execute(agent, args);
			
			IItem[] links = (IItem[]) response.getObject();
			IItem[] targets = new IItem[links.length];
			
			for(int i = 0; i < links.length; i++)
			{
				String path = links[i].getArg(Arg.TARGET);
				targets[i] = Server.instance.get(path);
			}
			
			String str = Model.instance.parser().composeItemArray(targets);
			
			// update the ACCESSED property of query results
			for(IItem item : targets)
			{
				item.access();
				Server.instance.put(item);
			}
			
			return new Response(ResponseType.ITEM_LIST, str);
		}
	},
	
	EMAIL
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String recipient = args.get(Arg.RECIPIENT);
			String subject = args.get(Arg.SUBJECT);
			String text = args.get(Arg.TEXT);
			
			boolean result = Server.instance.emailer().send(recipient, subject, text);
			
			return new Response(result, getNews(agent, args));
		}
	},
	
	REGISTER
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String email = args.get(Arg.EMAIL);
			String id = args.get(Arg.ID);
			String password = args.get(Arg.PASSWORD);
			String confirm = args.get(Arg.CONFIRM_PASSWORD);
			
			if(!password.equals(confirm))
				return new Response(false, "password and confirm password do not match");
			
			IItem user = new Item(
					Arg.TYPE.put(ItemType.USER),
					Arg.EMAIL.put(email), 
					Arg.ID.put(id), 
					Arg.PASSWORD.put(password));
			
			Server.instance.post(user);
			
			boolean result = Server.instance.get(user.getPath()) != null;
			
			return new Response(result, getNews(agent, args));
		}
	},
	
	CONFIRM_EMAIL
	{
		public IResponse execute(String agent, Map<IArg, String> args)
		{
			String id = args.get(Arg.ID);
			String confirm = args.get(Arg.CONFIRMATION_CODE);
			
			String path = ItemType.USER.getInstancePath(id);
			
			IItem user = Server.instance.get(path);
			
			if(user == null)
				return new Response(false, "user '" + id + "' not found");
			
			String encrypted = user.getArg(Arg.CONFIRMATION_CODE);
			
			boolean result = Server.instance.encryptor().check(confirm, encrypted);
			
			if(result)
			{
				user.setArg(Arg.EMAIL_CONFIRMED, Boolean.toString(true));
				Server.instance.put(user);
			}
			
			return new Response(result, getNews(agent, args));
		}
	}
	
	;
	
	public ICmd getCmd()
	{
		return Model.instance.getCmd(name());
	}
	
	public String getNews (String agent, Map<IArg, String> args)
	{
		return getCmd().getNews(agent, args);
	}
	
	public IResponse validate(String agent, Map<IArg, String> args)
	{
		return new Response(true);
	}
	
	public abstract IResponse execute(String agent, Map<IArg, String> args);

	public static ICmdImpl[] getValues()
	{
		return (ICmdImpl[])values();
	}
}
