package org.colcon.burocrat.server;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.colcon.burocrat.api.model.IArg;
import org.colcon.burocrat.api.model.ICmd;
import org.colcon.burocrat.api.model.ICommand;
import org.colcon.burocrat.api.model.IItem;
import org.colcon.burocrat.api.model.IResponse;
import org.colcon.burocrat.api.server.IArgImpl;
import org.colcon.burocrat.api.server.IAuthenticator;
import org.colcon.burocrat.api.server.ICmdImpl;
import org.colcon.burocrat.api.server.IDatabase;
import org.colcon.burocrat.api.server.IEmailer;
import org.colcon.burocrat.api.server.IEncryptor;
import org.colcon.burocrat.api.server.IItemTypeImpl;
import org.colcon.burocrat.api.server.INotifier;
import org.colcon.burocrat.api.server.IServer;
import org.colcon.burocrat.api.utils.EnumUtil;
import org.colcon.burocrat.api.utils.StringUtil;
import org.colcon.burocrat.model.Arg;
import org.colcon.burocrat.model.CmdFlag;
import org.colcon.burocrat.model.Item;
import org.colcon.burocrat.model.ItemType;
import org.colcon.burocrat.model.Model;
import org.colcon.burocrat.model.Response;

/**
 * 
 * The main server class
 * 
 * @author Dan Vegeto
 *
 */
public class Server implements IServer
{
	public static IServer instance = new Server
	(
		ArgImpl.values(), 
		CmdImpl.values(), 
		ItemTypeImpl.values(),
		
		new Database(), 
		new Authenticator(),
		new Encryptor(), 
		new Emailer(), 
		new Notifier()
	);
	
	public Map<String, IArgImpl> argImpls;
	public Map<String, ICmdImpl> cmdImpls;
	public Map<String, IItemTypeImpl> itemTypeImpls;
	
	public IDatabase database;
	public IAuthenticator authenticator;
	public IEncryptor encryptor;
	public IEmailer	emailer;
	public INotifier notifier;
	
	public Server(IArgImpl[] argImplsArray, ICmdImpl[] cmdImplsArray, IItemTypeImpl[] itemTypeImplsArray, 
			IDatabase database, IAuthenticator authenticator, IEncryptor encryptor, IEmailer emailer, INotifier notifier)
	{
		argImpls = new HashMap<String, IArgImpl>();
		cmdImpls = new HashMap<String, ICmdImpl>();
		itemTypeImpls = new HashMap<String, IItemTypeImpl>();
		
		EnumUtil.putAll(argImpls, argImplsArray);
		EnumUtil.putAll(cmdImpls, cmdImplsArray);
		EnumUtil.putAll(itemTypeImpls, itemTypeImplsArray);
		
		this.database = database;
		this.authenticator = authenticator;
		this.encryptor = encryptor;
		this.emailer = emailer;
		this.notifier = notifier;
	}
	
	public void init()
	{
		database().init();
		
		if(get(ItemType.METADATA.name()) == null)
		{
			for(IItem item : Model.instance.getInitItems())
				post(item);
		}
	}
	
	public IDatabase database() 
	{
		return database;
	}
	
	public IAuthenticator authenticator()
	{
		return authenticator;
	}
	
	public IEncryptor encryptor()
	{
		return encryptor;
	}
	
	public IEmailer	emailer()
	{
		return emailer;
	}
	
	public INotifier notifier()
	{
		return notifier;
	}

	public IArgImpl[] getArgImpls() 
	{
		return argImpls.values().toArray(new IArgImpl[0]);
	}
	
	public ICmdImpl[] getCmdImpls() 
	{
		return cmdImpls.values().toArray(new ICmdImpl[0]);
	}
	
	public IItemTypeImpl[] getItemTypeImpls() 
	{
		return itemTypeImpls.values().toArray(new IItemTypeImpl[0]);
	}

	public IArgImpl getArgImpl(String name) 
	{
		return argImpls.get(name.toUpperCase());
	}
	
	public ICmdImpl getCmdImpl(String name) 
	{
		return cmdImpls.get(name.toUpperCase());
	}
	
	public IItemTypeImpl getItemTypeImpl(String name) 
	{
		return itemTypeImpls.get(name.toUpperCase());
	}
	
	public IResponse validate(String agent, ICommand command)
	{
		if(command == null)
			return new Response(false, "invalid command");
		
		ICmd cmd = command.getCmd();
		
		if(cmd == null)
			return new Response(false, "invalid command");
		
		IResponse response = cmd.argList().validate(command.getArgs());
		
		if(!response.isSuccess())
			return response;
		
		response = cmd.validate(agent, command.getArgs());
		
		if(!response.isSuccess())
			return response;
		
		ICmdImpl cmdImpl = getCmdImpl(command.getCmd().name());
		
		return cmdImpl.validate(agent, command.getArgs());
	}

	public IResponse execute(String agent, ICommand command) 
	{
		IResponse response = validate(agent, command);
		
		if(!response.isSuccess())
			return response;
		
		ICmdImpl cmdImpl = getCmdImpl(command.getCmd().name());
		
		response = cmdImpl.validate(agent, command.getArgs());
		
		if(!response.isSuccess())
			return response;
		
		response = cmdImpl.execute(agent, command.getArgs());
		
		log(agent, command);
		
		return response;
	}
	
	public void log(String agent, ICommand command)
	{
		ICmd cmd = command.getCmd();

		// post news item
		String newsText = cmd.getNews(agent, command.getArgs());
		boolean isPublic = cmd.hasFlag(CmdFlag.PUBLIC);
		
		IItem news = new Item(
				Arg.TYPE.put(ItemType.NEWS),
				Arg.AGENT.put(agent), 
				Arg.COMMAND.put(command), 
				Arg.PUBLIC.put(isPublic));
		
		post(news);
		
		// post notification links
		String[] paths = StringUtil.getHashtagPaths(newsText);
		for(String path : paths)
		{
			IItem notification = new Item(
					Arg.TYPE.put(ItemType.NOTIFICATION),
					Arg.SOURCE.put(news.getPath()), 
					Arg.TARGET.put(path));
			
			post(notification);
			
			// notify (email, etc.)
			Server.instance.notifier().notify(path, newsText);
		}
	}

	public IItem get(String path) 
	{
		if(path == null)
			return null;
		
		Map<String, String> map = database().get(path);
		
		if(map == null)
			return null;
		
		IItem item = new Item(Model.instance.parser().stringsToArgs(map));
		
		put(item);
		
		return item;
	}

	public void post(IItem item) 
	{
		if(item == null || item.getType() == null)
			return;
		
		item.init();
		
		IItemTypeImpl typeImpl = getItemTypeImpl(item.getType().name());
		if(typeImpl == null)
			typeImpl = ItemTypeImpl.ITEM;
		
		typeImpl.init(item);
		
		Map<String, String> map = Model.instance.parser().argsToStrings(item.getArgs());
		
		database().post(map);
		
		for(IItem initItem : item.getInitItems())
			post(initItem);
	}

	public void put(IItem item) 
	{
		if(item == null)
			return;
		
		Map<String, String> map = Model.instance.parser().argsToStrings(item.getArgs());
		
		database().put(map);
	}

	public void delete(String path) 
	{
		if(path == null)
			return;
		
		database().delete(path);
	}

	public IItem[] query(String parent, int limit, int offset, IArg sort, boolean ascending, Map<IArg, String> filters) 
	{
		Map<String, String> filtersMap = Model.instance.parser().argsToStrings(filters);
		
		String sortStr = null;
		if(sort != null)
			sortStr = sort.name();
		
		List<Map<String, String>> list = database().query(parent, limit, offset, sortStr, ascending, filtersMap);
		
		IItem[] items = new IItem[list.size()];
		
		for(int i = 0; i < items.length; i++)
		{
			Map<IArg, String> map = Model.instance.parser().stringsToArgs(list.get(i));
			items[i] = new Item(map);
		}
		
		return items;
	}
}
